#include <cassert>
#include <stdio.h>
#include "exthashtbl.h"

#include <cstring>

#ifdef __BORLANDC__
#include <mem.h>	    // memcpy - grr
#endif


extern u8* mem_alloc(u32 size, bool in_idx);

// bits gets the hashkey and returns the k-th bit or the highest k bits - this is needed to order the keys
u32 ExtHashTbl::bits(u32 hashkey, u32 startpos, u32 length)   
{  
  
     u32 mask = 0x80000000;  
     if (!startpos && !length)   
     {  
          return 0;  
     } else if (length == 1) {  
          mask = mask >> (startpos);  
          mask = mask & hashkey;  
          mask >>= (31 - startpos);  
     } else {  
        mask >>= (length-1);  
          mask = mask - 1;  
          mask = ~mask;  
          mask = mask & hashkey;  
          mask = mask >>(32 - length);  
     };  
     return mask;  
};


// creating a new word/page
link ExtHashTbl::NEW() {
assert(sizeof(Word) == 12);
assert(sizeof(STnode) == 512);

	 // force it to be in the mapped memory area
     link x = (link) mem_alloc(sizeof(STnode), true);
     x->m = 0;
     x->k = 0;
     return x;
}; 

   
    
// search a page for hash key v
Item* ExtHashTbl::search_in_db(link h, u32 v) {
     for (int j = 0; j < h->m; j++)    
		 if (v == h->b[j].key)
               return &h->b[j];    

	return 0;
	
};    
    

// this function inserts the newly created page of a split page into the directory
void ExtHashTbl::insertDir(link t, int k) {    
     int i, m, x = bits(t->b[0].key, 0, k);    
     // double the old directory and copy the old entries into the new - as often as necessary
	 while (d < k) {    
          link *old = dir;    
          d += 1; D+= D;    
          dir = (link*)mem_alloc(D*sizeof(link), true);
          for (i = 0; i < D; i++)    
               dir[i] = old [i/2];    
          if (d < k)    
               dir[bits(x, 0, d) ^ 1] = NEW();    
     }    

	 // enter the newly created page into directory
	 for(m = 1; k < d; k++)
		 m *= 2;
	 for(i = 0; i < m; i++)
		 dir[x*m + i] = t;
};    

// h is full with keys. this function splits h
Item* ExtHashTbl::split(link h, int added) {      
      link t = NEW();  
      Item* new_entry;  
     while (h->m == 0 || h->m == M) {      
          h->m = 0;      
          t->m = 0;  
                  
          for (int j = 0; j < M; j++)  
				 // is the k-th bit = 1 => key is to stay in h
                 if (bits(h->b[j].key, h->k, 1) == 0) {       
                      h->b[(h->m)++] = h->b[j];  
                      if (j == added)  
                           new_entry = &h->b[h->m];  
                 } else {  
					 // the k-th bit = 0 => key is copied to t
                      t->b[(t->m)++] = h->b[j];  
                      if (j == added)  
                           new_entry = &t->b[t->m];  
                 }
           t->k = ++(h->k); 
		   // increase k - the number of equal leading bits 
		   // do this procedure until h is no longer completely filled with keys
     }  
	  // insert into directory 
      insertDir(t, t->k);  
	  // return the pointer to the item that was inserted 
      return new_entry;  
};   

// this function sorts the key that is to be inserted into the existing page and calls for a split if necessary
Item* ExtHashTbl::insert_in_db(link h, Item item) {
int j;
	Item* new_entry;
	// sort new key into existing page
	for (j = 0; j < h->m; j++)
          if (item.key < h->b[j].key)
			  break;
     for (int i = (h->m)++; i > j; i--)
          h->b[i] = h->b[i-1];    
     h->b[j] = item;    
	 // is a split needed?
	 if (h->m == M)
		 new_entry = split(h, j);
	 else
		 new_entry = &h->b[j];
	 // return the pointer to the item that was inserted
	 return new_entry;
};

void ExtHashTbl::init() {    
     d = 0;
     D = 1;    
     dir = (link*)mem_alloc(D*sizeof(dir), true);    
     dir[0] = NEW();    
}; 

// get the directory entry (the highest d bits) and start the inserting function on that page 
Item* ExtHashTbl::insert(Item& item) {
	int idx = bits(item.key, 0, d);
	assert(idx < D);
	return insert_in_db(dir[idx], item);
};

// get the directory entry (the highest d bits) and start the search function on that page
Item* ExtHashTbl::search(u32 v) {
	int idx = bits(v, 0, d);
	assert(idx < D);
	return search_in_db(dir[idx], v);
};

// this function fixes the pointers in the reloaded mapped index file
void ExtHashTbl::rebase(long delta){

	// note! can't rebase dir in place, b/c idx mapping is read-only!

	int dir_size = D*sizeof(STnode*);
	(long&)dir += delta;
	STnode** new_dir = (STnode**)mem_alloc(dir_size, false);
	memcpy(new_dir, dir, dir_size);

	for(int i = 0; i < D; i++)
		(long&)new_dir[i] += delta;

	dir = new_dir;
}

// used for testing...
void ExtHashTbl::tree_out() {
/*
	for(int i = 0; i < D; i++) {
		printf("dirent %d, bits %x -> page# %d: ", i, bits(dir[i]->b[0].key, 0, d), dir[i]->pagenum);
		for (int j = 0; j < dir[i]->m; j++) 
			printf("%x, ", dir[i]->b[j].value);
		printf("\n");   
	};
	printf("\n\n");
*/
}; 
    





