//#define DUMP_AR

#define u8 unsigned char
#define u16 unsigned short
#define u32 unsigned long
#define u64 unsigned __int64
#define uint unsigned int

// tunables

const int MIN_WORD_LEN = 4;		// words shorter than this aren't indexed
const int MAX_WORD_LEN = 31;	// max length, due to fixed size buffer
								//TODO: how to cut off?

const int MAX_OCCS = 200;		// word occurs more often => it's made a stopword

const int BUCKET_ENTRIES = 7;	// so bucket fits in a cache line
								// tradeoff between internal frag, and overhead
const int BUCKET_SIZE = (BUCKET_ENTRIES+1)*4;

const u32 BLOCK_SIZE = 32*1024;
const int MAX_ACTIVE_READS = 4;

// Occurrence field widths. some code needs these limits
const int FILE_NUM_BITS = 16;
const int MAX_FILES = 1 << FILE_NUM_BITS;

const int BLOCK_BITS = 8;
const int MAX_BLOCK = (1 << BLOCK_BITS) - 1;



#ifdef __BORLANDC__
#define strnicmp strnicmp
#elif defined(_MSC_VER)
#define strnicmp _strnicmp
#else
#error "port strnicmp"
#endif


//------------------------------ detect.cpp ------------------------------//

enum FileType
{
	// type < 0 <==> file needs to be read in one piece
	FT_PDF = -1,			// includes .ps
	FT_ZIP = -2,
	FT_RAR = -3,

	// the file is of a known format that does not contain any text
	// read code requires this to be 0
	FT_NON_TEXT = 0,

	// initial file type - we don't yet know the format.
	// only returned if determining the format requires > 1 block.
	FT_UNKNOWN,

	// formats that are parsed differently
	FT_HTML,
	FT_MP3,
	FT_ISO,

	// anything else. call contains_text for each block.
	FT_POSSIBLE_TEXT
};


// called for every block read from files of type FT_UNKNOWN.
extern FileType detect_file_type(const u8* buf, const int size);



// check if the specified buffer contains reasonable text.
//
// output:
// - start_ofs and end_ofs are optional.
//   if true is returned, and the pointers are valid, they receive the maximum
//   extent of text in the buffer.
extern bool contains_text(const u8* buf, const int size, int* start_ofs, int* end_ofs);



//------------------------------ stem.cpp ------------------------------//

// reduces word to its stem (in place).
// TODO: return value indicates if changes were made.?
extern void get_word_stem(char* word, int& len);


//------------------------------ read.cpp ------------------------------//

extern long cur_fn;
extern const char* cur_path;
extern unsigned cur_block;

extern bool read_cancelled;

extern int gen_file_list(const char* path);
extern int read_files_in_list();

extern const char* get_filename(u32 file_num);

extern long read_block(u32 file_num, u32 block, const u8*& buf, long& size);

// read progress, used by frontend
extern long num_files;
extern u64 total_bytes, total_bytes_read;
extern float read_throughput;	// [byte/s]


//------------------------------ extract.cpp ------------------------------//

extern long extract_words(const u8* buf, u32 size);
extern long process_buf(const u8* buf, u32 size, FileType type);


//------------------------------ idx.cpp ------------------------------//

extern u64 total_text, total_words, unique_words;


// word mem may be modified, e.g. by stemmer
extern int lookup_word(char* word, int len);

extern long create_idx(const char* path);
extern int cancel_idx();
extern int read_idx();

enum IdxStatus
{
	IS_INACTIVE,
	IS_ACTIVE,
	IS_SUCCEEDED,
	IS_FAILED
};
extern IdxStatus idx_status();

extern long search(const char* string);
extern int get_next_search_result(const char*& filename, int& freq, const char** context);


extern u8* mem_alloc(u32 size, bool in_idx);
extern void mem_free(u8* p, u32 size);
extern void mem_reset();

