#include <cassert>
#include <stdlib.h>
#include <cmath>

#include "time.h"
#include "misc.h"


#include <map>




u16 bswap16(u16 x)
{
	return ((x & 0xff) << 8) | (x >> 8);
}

u32 bswap32(u32 x)
{
#ifdef _M_IX86

__asm
{
	mov		eax, [x]
	bswap	eax
	mov		[x], eax
}

#else

	u32 t = x;

	for(int i = 0; i < 4; i++)
	{
		x <<= 8;
		x |= t & 0xff;
	}

#endif
	return x;
}


bool is_pow2(const int n)
{
	return (n > 0) && !(n & (n-1));
}


/*
 * return -1 if not an integral power of 2,
 *        otherwise the base2 logarithm
 */
int log2(const int n)
{
#ifdef _M_IX86

	__asm
	{
		mov		ecx, [n]
		or		eax, -1			// return value
		lea		edx, [ecx-1]
		test	ecx, edx		// power of 2?
		jnz		$ret
		bsf		eax, ecx
	$ret:
		mov		[n], eax
	}

	return n;

#else

	if(n || n & (n-1))
		return -1;

	int i = 1, j = 0;
	for(; i != n; i += i, j++)
		;
	return j;

#endif
}


static int log2(const float x)
{
	u32 i = (u32&)x;
	u32 exp = (i >> 23) & 0xff;
	return (int)exp - 127;
}


long round_up(long val, int multiple)
{
	val += multiple-1;
	val -= val % multiple;
	return val;
}


int ceil(float f)
{
	int r;

#ifdef _M_IX86
	const float _49 = 0.499999f;
	__asm
	{
		fld			[f]
		fadd		[_49]
		fistp		[r]
	}
#else
	r = (int)(f+0.999999f);
#endif

	return r;
}




void bswap32(const u8* data, int cnt)
{
	__asm
	{
		mov		edx, [data]
		mov		ecx, [cnt]
$loop:	dec		ecx
		js		$ret
		mov		eax, [edx+ecx*4]
		bswap	eax
		mov		[edx+ecx*4], eax
		jmp		$loop
$ret:
	}
}


/* big endian! */
void base32(const int len, const u8* in, u8* out)
{
	int bits = 0;
	u32 pool = 0;

	static u8 tbl[33] = "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567";

	for(int i = 0; i < len; i++)
	{
		if(bits < 5)
		{
			pool <<= 8;
			pool |= *in++;
			bits += 8;
		}

		bits -= 5;
		int c = (pool >> bits) & 31;
		*out++ = tbl[c];
	}
}

/*
typedef std::map<void*, void*> allocs_map;	// aligned address -> malloc address
typedef allocs_map::iterator allocs_it;
static allocs_map allocs;

static void alloc_cleanup(void)
{
	allocs_it it;
	for(it = allocs.begin(); it != allocs.end(); ++it)
		free((*it).second);
}


void _free(void* p)
{
	allocs_it it = allocs.find(p);
	if(it == allocs.end())
		return;

	free((*it).second);
	allocs.erase(it);
}
*/

void* alloc(int size, int align)
{
	void* p = malloc(size);
	if(!p)
		exit(1);

	// align
	void* aligned_p = (void*)round_up((int)p, align);
/*
	allocs[aligned_p] = p;
*/
	return aligned_p;
}





/*
void* large_malloc(int size)
{
	return VirtualAlloc(0, size, PAGE_COMMIT, PAGE_READWRITE);
}


void large_free(void* p)
{
	VirtualFree(p, 0, MEM_RELEASE);
}
*/